REM
REM  Joypad emulation of a GameBoy console
REM  Part of the zBoy project
REM  Copyright (C) Mateusz Viste 2010
REM

REM The joypad is controlled via the register at $FF00:
REM  Bit 5 - P15 out
REM  Bit 4 - P14 out
REM  Bit 3 - P13 in
REM  Bit 2 - P12 in
REM  Bit 1 - P11 in
REM  Bit 0 - P10 in
REM
REM  When one of the keys gets a transition from high to low, an interrupt is fired.

TYPE KeyDefinition ' Used by keymap definitions
  Type AS UBYTE  ' 0 = keyboard (multikey)  ;  1 = joypad (getjoystick)
  Key AS INTEGER
  JoyId AS INTEGER
  JoyButMask AS INTEGER  ' button mask, eg. &b0000000000001000 for button #4
  Axis AS UBYTE   ' axis number (1..4)
  AxisSig AS BYTE ' axis signum (-1 or +1)
END TYPE

TYPE KeymapDefinition ' Used for keymap definition
  Up AS KeyDefinition
  Down AS KeyDefinition
  Left AS KeyDefinition
  Right AS KeyDefinition
  Start AS KeyDefinition
  Select AS KeyDefinition
  A AS KeyDefinition
  B AS KeyDefinition
  TurboA AS KeyDefinition
  TurboB AS KeyDefinition
  bgCtrl AS KeyDefinition
  spriteCtrl AS KeyDefinition
  Load AS KeyDefinition
  Save AS KeyDefinition
  Autoshot AS KeyDefinition
  Screenshot AS KeyDefinition
  Reset AS KeyDefinition
  Quit AS KeyDefinition
END TYPE

TYPE KeyStateMap  ' Used for internal monitoring of key states
  Up AS BYTE
  Down AS BYTE
  Right AS BYTE
  Left AS BYTE
  Select AS BYTE
  Start AS BYTE
  A AS BYTE
  B AS BYTE
  F1 AS BYTE
  F2 AS BYTE
  F3 AS BYTE
  F4 AS BYTE
  F5 AS BYTE
  F6 AS BYTE
  F7 AS BYTE
  F8 AS BYTE
  F9 AS BYTE
  F10 AS BYTE
  ESC AS BYTE
END TYPE

DIM SHARED JoypadAxis(1 TO 8) AS SINGLE  ' axis (1..4 for first joypad, 5..8 for second)
DIM SHARED JoypadButtons(0 TO 1) AS INTEGER

DIM SHARED AS ULONGINT TurboTimerA, TurboTimerB

DIM SHARED AS UBYTE JoyOldReg = 0, JoyNewReg, AskForScreenshot = 0
DIM SHARED AS UBYTE ThereIsAJoypad = 0, ThereIsASecondJoypad = 0
DIM SHARED AS ULONGINT JoyNextCheck
DIM SHARED KeybState AS KeyStateMap
DIM SHARED KeyDef AS KeymapDefinition


FUNCTION zboyScancodeToFbScancode(zboyScanCode AS STRING, KeydefVar AS KeyDefinition) AS UBYTE
  SELECT CASE LCASE(zboyScanCode)
    CASE "none"   '  No key assigned
      KeydefVar.Type = 255
    CASE "kb.f1"
      KeydefVar.Type = 0
      KeydefVar.key = &h3B
    CASE "kb.f2"
      KeydefVar.Type = 0
      KeydefVar.key = &h3C
    CASE "kb.f3"
      KeydefVar.Type = 0
      KeydefVar.key = &h3D
    CASE "kb.f4"
      KeydefVar.Type = 0
      KeydefVar.key = &h3E
    CASE "kb.f5"
      KeydefVar.Type = 0
      KeydefVar.key = &h3F
    CASE "kb.f6"
      KeydefVar.Type = 0
      KeydefVar.key = &h40
    CASE "kb.f7"
      KeydefVar.Type = 0
      KeydefVar.key = &h41
    CASE "kb.f8"
      KeydefVar.Type = 0
      KeydefVar.key = &h42
    CASE "kb.f9"
      KeydefVar.Type = 0
      KeydefVar.key = &h43
    CASE "kb.f10"
      KeydefVar.Type = 0
      KeydefVar.key = &h44
    CASE "kb.f11"
      KeydefVar.Type = 0
      KeydefVar.key = &h57
    CASE "kb.f12"
      KeydefVar.Type = 0
      KeydefVar.key = &h58
    CASE "kb.esc"
      KeydefVar.Type = 0
      KeydefVar.key = &h01
    CASE "kb.tab"
      KeydefVar.Type = 0
      KeydefVar.key = &h0F
    CASE "kb.ctrl"
      KeydefVar.Type = 0
      KeydefVar.key = &h1D
    CASE "kb.alt"
      KeydefVar.Type = 0
      KeydefVar.key = &h38
    CASE "kb.spc"
      KeydefVar.Type = 0
      KeydefVar.key = &h39
    CASE "kb.ret"
      KeydefVar.Type = 0
      KeydefVar.key = &h1C
    CASE "kb.bks"
      KeydefVar.Type = 0
      KeydefVar.key = &h0E
    CASE "kb.left"
      KeydefVar.Type = 0
      KeydefVar.key = &h4B
    CASE "kb.right"
      KeydefVar.Type = 0
      KeydefVar.key = &h4D
    CASE "kb.up"
      KeydefVar.Type = 0
      KeydefVar.key = &h48
    CASE "kb.down"
      KeydefVar.Type = 0
      KeydefVar.key = &h50
    CASE "kb.rshift"
      KeydefVar.Type = 0
      KeydefVar.key = &h36
    CASE "kb.lshift"
      KeydefVar.Type = 0
      KeydefVar.key = &h2A
    CASE "kb.lwin"
      KeydefVar.Type = 0
      KeydefVar.key = &h7D
    CASE "kb.rwin"
      KeydefVar.Type = 0
      KeydefVar.key = &h7E
    CASE "kb.menu"
      KeydefVar.Type = 0
      KeydefVar.key = &h7F
    CASE "kb.pgup"
      KeydefVar.Type = 0
      KeydefVar.key = &h49
    CASE "kb.pgdw"
      KeydefVar.Type = 0
      KeydefVar.key = &h51
    CASE "kb.home"
      KeydefVar.Type = 0
      KeydefVar.key = &h47
    CASE "kb.end"
      KeydefVar.Type = 0
      KeydefVar.key = &h4F
    CASE "kb.ins"
      KeydefVar.Type = 0
      KeydefVar.key = &h52
    CASE "kb.del"
      KeydefVar.Type = 0
      KeydefVar.key = &h53
    CASE "kb.a"
      KeydefVar.Type = 0
      KeydefVar.key = &h1E
    CASE "kb.b"
      KeydefVar.Type = 0
      KeydefVar.key = &h30
    CASE "kb.c"
      KeydefVar.Type = 0
      KeydefVar.key = &h2E
    CASE "kb.d"
      KeydefVar.Type = 0
      KeydefVar.key = &h20
    CASE "kb.e"
      KeydefVar.Type = 0
      KeydefVar.key = &h12
    CASE "kb.f"
      KeydefVar.Type = 0
      KeydefVar.key = &h21
    CASE "kb.g"
      KeydefVar.Type = 0
      KeydefVar.key = &h22
    CASE "kb.h"
      KeydefVar.Type = 0
      KeydefVar.key = &h23
    CASE "kb.i"
      KeydefVar.Type = 0
      KeydefVar.key = &h17
    CASE "kb.j"
      KeydefVar.Type = 0
      KeydefVar.key = &h24
    CASE "kb.k"
      KeydefVar.Type = 0
      KeydefVar.key = &h25
    CASE "kb.l"
      KeydefVar.Type = 0
      KeydefVar.key = &h26
    CASE "kb.m"
      KeydefVar.Type = 0
      KeydefVar.key = &h32
    CASE "kb.n"
      KeydefVar.Type = 0
      KeydefVar.key = &h31
    CASE "kb.o"
      KeydefVar.Type = 0
      KeydefVar.key = &h18
    CASE "kb.p"
      KeydefVar.Type = 0
      KeydefVar.key = &h19
    CASE "kb.q"
      KeydefVar.Type = 0
      KeydefVar.key = &h10
    CASE "kb.r"
      KeydefVar.Type = 0
      KeydefVar.key = &h13
    CASE "kb.s"
      KeydefVar.Type = 0
      KeydefVar.key = &h1F
    CASE "kb.t"
      KeydefVar.Type = 0
      KeydefVar.key = &h14
    CASE "kb.u"
      KeydefVar.Type = 0
      KeydefVar.key = &h16
    CASE "kb.v"
      KeydefVar.Type = 0
      KeydefVar.key = &h2F
    CASE "kb.w"
      KeydefVar.Type = 0
      KeydefVar.key = &h11
    CASE "kb.x"
      KeydefVar.Type = 0
      KeydefVar.key = &h2D
    CASE "kb.y"
      KeydefVar.Type = 0
      KeydefVar.key = &h15
    CASE "kb.z"
      KeydefVar.Type = 0
      KeydefVar.key = &h2C
    CASE "kb.0"
      KeydefVar.Type = 0
      KeydefVar.key = &h0B
    CASE "kb.1"
      KeydefVar.Type = 0
      KeydefVar.key = &h02
    CASE "kb.2"
      KeydefVar.Type = 0
      KeydefVar.key = &h03
    CASE "kb.3"
      KeydefVar.Type = 0
      KeydefVar.key = &h04
    CASE "kb.4"
      KeydefVar.Type = 0
      KeydefVar.key = &h05
    CASE "kb.5"
      KeydefVar.Type = 0
      KeydefVar.key = &h06
    CASE "kb.6"
      KeydefVar.Type = 0
      KeydefVar.key = &h07
    CASE "kb.7"
      KeydefVar.Type = 0
      KeydefVar.key = &h08
    CASE "kb.8"
      KeydefVar.Type = 0
      KeydefVar.key = &h09
    CASE "kb.9"
      KeydefVar.Type = 0
      KeydefVar.key = &h0A
    CASE "joy1.a1pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 1
      KeydefVar.AxisSig = 1
      ThereIsAJoypad = 1
    CASE "joy1.a1neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 1
      KeydefVar.AxisSig = -1
      ThereIsAJoypad = 1
    CASE "joy1.a2pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 2
      KeydefVar.AxisSig = 1
      ThereIsAJoypad = 1
    CASE "joy1.a2neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 2
      KeydefVar.AxisSig = -1
      ThereIsAJoypad = 1
    CASE "joy1.a3pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 3
      KeydefVar.AxisSig = 1
      ThereIsAJoypad = 1
    CASE "joy1.a3neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 3
      KeydefVar.AxisSig = -1
      ThereIsAJoypad = 1
    CASE "joy1.a4pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 4
      KeydefVar.AxisSig = 1
      ThereIsAJoypad = 1
    CASE "joy1.a4neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.Axis = 4
      KeydefVar.AxisSig = -1
      ThereIsAJoypad = 1
    CASE "joy1.b1"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000000000001
      ThereIsAJoypad = 1
    CASE "joy1.b2"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000000000010
      ThereIsAJoypad = 1
    CASE "joy1.b3"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000000000100
      ThereIsAJoypad = 1
    CASE "joy1.b4"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000000001000
      ThereIsAJoypad = 1
    CASE "joy1.b5"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000000010000
      ThereIsAJoypad = 1
    CASE "joy1.b6"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000000100000
      ThereIsAJoypad = 1
    CASE "joy1.b7"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000001000000
      ThereIsAJoypad = 1
    CASE "joy1.b8"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000010000000
      ThereIsAJoypad = 1
    CASE "joy1.b9"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000000100000000
      ThereIsAJoypad = 1
    CASE "joy1.b10"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000001000000000
      ThereIsAJoypad = 1
    CASE "joy1.b11"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000010000000000
      ThereIsAJoypad = 1
    CASE "joy1.b12"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0000100000000000
      ThereIsAJoypad = 1
    CASE "joy1.b13"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0001000000000000
      ThereIsAJoypad = 1
    CASE "joy1.b14"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0010000000000000
      ThereIsAJoypad = 1
    CASE "joy1.b15"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b0100000000000000
      ThereIsAJoypad = 1
    CASE "joy1.b16"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 0
      KeydefVar.JoyButMask = &b1000000000000000
      ThereIsAJoypad = 1
    CASE "joy2.a1pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 1
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a1neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 1
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.a2pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 2
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a2neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 2
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.a3pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 3
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a3neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 3
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.a4pos"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 4
      KeydefVar.AxisSig = 1
      ThereIsASecondJoypad = 1
    CASE "joy2.a4neg"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.Axis = 4
      KeydefVar.AxisSig = -1
      ThereIsASecondJoypad = 1
    CASE "joy2.b1"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000000001
      ThereIsASecondJoypad = 1
    CASE "joy2.b2"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000000010
      ThereIsASecondJoypad = 1
    CASE "joy2.b3"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000000100
      ThereIsASecondJoypad = 1
    CASE "joy2.b4"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000001000
      ThereIsASecondJoypad = 1
    CASE "joy2.b5"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000010000
      ThereIsASecondJoypad = 1
    CASE "joy2.b6"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000000100000
      ThereIsASecondJoypad = 1
    CASE "joy2.b7"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000001000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b8"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000010000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b9"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000000100000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b10"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000001000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b11"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000010000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b12"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0000100000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b13"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0001000000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b14"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0010000000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b15"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b0100000000000000
      ThereIsASecondJoypad = 1
    CASE "joy2.b16"
      KeydefVar.Type = 1
      KeydefVar.JoyId = 1
      KeydefVar.JoyButMask = &b1000000000000000
      ThereIsASecondJoypad = 1
    CASE ELSE
      RETURN 1
  END SELECT
END FUNCTION


FUNCTION RedefineInputKeys(sKeyUp AS STRING, sKeyDown AS STRING, sKeyLeft AS STRING, sKeyRight AS STRING, sKeyStart AS STRING, sKeySelect AS STRING, sKeyA AS STRING, sKeyB AS STRING, sKeyTurboA AS STRING, sKeyTurboB AS STRING) AS UBYTE
  IF zboyScancodeToFbScancode(sKeyUp, KeyDef.Up) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyDown, KeyDef.Down) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyLeft, KeyDef.Left) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyRight, KeyDef.Right) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyStart, KeyDef.Start) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeySelect, KeyDef.Select) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyA, KeyDef.A) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyB, KeyDef.B) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyTurboA, KeyDef.TurboA) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyTurboB, KeyDef.TurboB) <> 0 THEN RETURN 1
  RETURN 0
END FUNCTION


FUNCTION RedefineFunctionKeys(sKeyBg AS STRING, sKeySprite AS STRING, sKeySave AS STRING, sKeyLoad AS STRING, sKeyAutoshot AS STRING, sKeyShot AS STRING, sKeyRst AS STRING, sKeyQuit AS STRING) AS UBYTE
  IF zboyScancodeToFbScancode(sKeyBg, KeyDef.BgCtrl) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeySprite, KeyDef.SpriteCtrl) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeySave, KeyDef.Save) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyLoad, KeyDef.Load) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyAutoshot, KeyDef.Autoshot) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyShot, KeyDef.Screenshot) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyRst, KeyDef.Reset) <> 0 THEN RETURN 1
  IF zboyScancodeToFbScancode(sKeyQuit, KeyDef.Quit) <> 0 THEN RETURN 1
  RETURN 0
END FUNCTION


FUNCTION CheckKeyState(BYREF KeyToPoll AS KeyDefinition) AS UBYTE ' returns 1 if key is pressed, 0 otherwise
  SELECT CASE KeyToPoll.Type
    CASE 0  ' Keyboard
      IF MULTIKEY(KeyToPoll.Key) = -1 THEN
          RETURN 1
        ELSE
          RETURN 0
      END IF
    CASE 1  ' Joypad
      IF KeyToPoll.JoyButMask > 0 THEN  ' The thing to poll is a button
          IF (JoypadButtons(KeyToPoll.JoyId) AND KeyToPoll.JoyButMask) > 0 THEN
              RETURN 1
            ELSE
              RETURN 0
          END IF
        ELSE   ' The thing to poll is an axis
          IF KeyToPoll.AxisSig = 1 THEN  ' Positive axis
              IF JoypadAxis((KeyToPoll.JoyId SHL 2) + KeyToPoll.Axis) > (0.30) THEN
                  RETURN 1
                ELSE
                  RETURN 0
              END IF
            ELSE  ' Negative axis
              IF JoypadAxis((KeyToPoll.JoyId SHL 2) + KeyToPoll.Axis) < (-0.30) THEN
                  RETURN 1
                ELSE
                  RETURN 0
              END IF
          END IF
      END IF
    CASE 255  ' No key binding assigned, always return "0" (not pressed)
      RETURN 0
  END SELECT
END FUNCTION


#MACRO CallJoyINT() ' Request a joypad INT
  IoRegisters(&hFF0F) = (IoRegisters(&hFF0F) OR &b00010000) ' Request a joypad INT
#ENDMACRO

#MACRO DownIsPressed()
  IF (KeybState.Down = 0) THEN
    KeybState.Down = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO

#MACRO UpIsPressed()
  IF (KeybState.Up = 0) THEN
    KeybState.Up = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO

#MACRO LeftIsPressed()
  IF (KeybState.Left = 0) THEN
    KeybState.Left = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO

#MACRO RightIsPressed()
  IF (KeybState.Right = 0) THEN
    KeybState.Right = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO

#MACRO StartIsPressed()
  IF (KeybState.Start = 0) THEN
    KeybState.Start = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO

#MACRO SelectIsPressed()
  IF (KeybState.Select = 0) THEN
    KeybState.Select = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO

#MACRO ButtonAisPressed()
  IF (KeybState.A = 0) THEN
    KeybState.A = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO

#MACRO ButtonBisPressed()
  IF (KeybState.B = 0) THEN
    KeybState.B = 1
    CallJoyINT() ' Request a joypad INT
  END IF
#ENDMACRO


#MACRO CheckJoypad() ' Is it *really* faster to use a MACRO here instead of SUB? I couldn't get any definitive test results... :-/
  IF (TotalCycles > JoyNextCheck) THEN
    'JoyNextCheck += 10000   ' Check joypad state every ~2.5 ms
    'JoyNextCheck += 20000   ' Check joypad state every ~5 ms
    'JoyNextCheck += 40000   ' Check joypad state every ~10 ms
    'JoyNextCheck += 60000   ' Check joypad state every ~15 ms
    'JoyNextCheck += 80000   ' Check joypad state every ~20 ms
    JoyNextCheck += 160000   ' Check joypad state every ~40 ms
    'JoyNextCheck += 320000   ' Check joypad state every ~80 ms

    IF (ThereIsAJoypad = 1) THEN   ' If Joypad #1 is used, poll it now
      GetJoystick(0, JoypadButtons(0), JoypadAxis(1), JoypadAxis(2), JoypadAxis(3), JoypadAxis(4))
    END IF
    IF (ThereIsASecondJoypad = 1) THEN   ' If Joypad #2 is used, poll it now
      GetJoystick(1, JoypadButtons(1), JoypadAxis(5), JoypadAxis(6), JoypadAxis(7), JoypadAxis(8))
    END IF

    REM Check extended emulator keys (snapshot, save, load, etc...)
    IF (CheckKeyState(KeyDef.BgCtrl) = 1) THEN  ' ctrlBG pressed
        IF (KeybState.F1 = 0) THEN
          KeybState.F1 = 1
        END IF
      ELSE
        IF (KeybState.F1 = 1) THEN
          IF (HideBackgroundDisplay = 1) THEN
              HideBackgroundDisplay = 0
              UserMessage = "BACKGROUND ON"
            ELSE
              HideBackgroundDisplay = 1
              UserMessage = "BACKGROUND OFF"
          END IF
          KeybState.F1 = 0
        END IF
    END IF
    IF (CheckKeyState(KeyDef.spriteCtrl) = 1) THEN  ' ctrlSprite pressed
        IF (KeybState.F2 = 0) THEN
          KeybState.F2 = 1
        END IF
      ELSE
        IF (KeybState.F2 = 1) THEN
          IF HideSpritesDisplay = 1 THEN
              HideSpritesDisplay = 0
              UserMessage = "SPRITES ON"
            ELSE
              HideSpritesDisplay = 1
              UserMessage = "SPRITES OFF"
          END IF
          KeybState.F2 = 0
        END IF
    END IF
    IF (CheckKeyState(KeyDef.Save) = 1) THEN  ' (save game pressed)
        IF (KeybState.F5 = 0) THEN
          KeybState.F5 = 1
        END IF
      ELSE
        IF (KeybState.F5 = 1) THEN
          KeybState.F5 = 0
          SaveGame()
        END IF
    END IF
    IF (CheckKeyState(KeyDef.Load) = 1) THEN  ' (load game pressed)
        IF (KeybState.F7 = 0) THEN
          KeybState.F7 = 1
        END IF
      ELSE
        IF (KeybState.F7 = 1) THEN
          KeybState.F7 = 0
          LoadGame()
        END IF
    END IF
    IF (CheckKeyState(KeyDef.Autoshot) = 1) THEN  ' (autoscreenshot pressed)
        IF (KeybState.F8 = 0) THEN
          KeybState.F8 = 1
        END IF
      ELSE
        IF (KeybState.F8 = 1) THEN
          KeybState.F8 = 0
          AutoScreenshot = 1 - AutoScreenshot   ' switch Autoscreenshot mode on/off
          IF (AutoScreenshot = 0) THEN
              UserMessage = "AUTOSCREENSHOT OFF"
            ELSE
              UserMessage = "AUTOSCREENSHOT ON"
              LastScreenshotTime = TIMER  ' avoid an immediate screenshot
          END IF
        END IF
    END IF
    IF (CheckKeyState(KeyDef.Screenshot) = 1) THEN  ' Screenshot key pressed
        IF (KeybState.F9 = 0) THEN
          KeybState.F9 = 1
        END IF
      ELSE
        IF (KeybState.F9 = 1) THEN
          AskForScreenshot = 1
          KeybState.F9 = 0
        END IF
    END IF
    IF (CheckKeyState(KeyDef.Reset) = 1) THEN  ' Reset key pressed
        IF (KeybState.F10 = 0) THEN
          KeybState.F10 = 1
        END IF
      ELSE
        IF (KeybState.F10 = 1) THEN
          PrintMsg("RESET...", 1)
          'RefreshScreen()    ' do not use refreshscreen (already called by PrintMsg)
          SLEEP 400, 1
          IF (HiScoresMem = 1) THEN
            SaveHiScoresMem()
          END IF
          ResetEmulator()   ' reset the emulator
          IF (HiScoresMem = 1) THEN ' Check out if need to look after hiscores
            LoadHiScoresMem()
          END IF
          KeybState.F10 = 0
        END IF
    END IF
    IF (CheckKeyState(KeyDef.Quit) = 1) THEN  ' Quit pressed
        IF (KeybState.ESC = 0) THEN
          KeybState.ESC = 1
        END IF
      ELSE
        IF (KeybState.ESC = 1) THEN
          QuitEmulator = 1
          KeybState.ESC = 0
        END IF
    END IF


    REM Poll the standard keyboard keys
    IF (CheckKeyState(KeyDef.Down) = 1) THEN  ' Down
        DownIsPressed()
      ELSE
        KeybState.Down = 0
    END IF
    IF (CheckKeyState(KeyDef.Up) = 1) THEN ' Up
        UpIsPressed()
      ELSE
        KeybState.Up = 0
    END IF
    IF (CheckKeyState(KeyDef.Left) = 1) THEN ' Left
        LeftIsPressed()
      ELSE
        KeybState.Left = 0
    END IF
    IF (CheckKeyState(KeyDef.Right) = 1) THEN ' Right
        RightIsPressed()
      ELSE
        KeybState.Right = 0
    END IF
    IF (CheckKeyState(KeyDef.Start) = 1) THEN ' Start
        StartIsPressed()
      ELSE
        KeybState.Start = 0
    END IF
    IF (CheckKeyState(KeyDef.Select) = 1) THEN ' Select
        SelectIsPressed()
      ELSE
        KeybState.Select = 0
    END IF
    IF (CheckKeyState(KeyDef.TurboA) = 1) THEN ' Button Turbo A
        IF TotalCycles > TurboTimerA THEN
          IF KeybState.A = 0 THEN
              ButtonAisPressed()
              TurboTimerA = TotalCycles + 400000   ' maintain button pressed for ~1/10s
            ELSE
              KeybState.A = 0
              TurboTimerA = TotalCycles + 800000  ' keep button released for ~1/5s
          END IF
        END IF
      ELSE  ' If turbo A not pressed, check out the regular A button...
        IF (CheckKeyState(KeyDef.A) = 1) THEN ' Button A
            ButtonAisPressed()
          ELSE
            KeybState.A = 0
        END IF
    END IF
    IF (CheckKeyState(KeyDef.TurboB) = 1) THEN ' Button Turbo B
        IF TotalCycles > TurboTimerB THEN
          IF KeybState.B = 0 THEN
              ButtonBisPressed()
              TurboTimerB = TotalCycles + 400000   ' maintain button pressed for ~1/10s
            ELSE
              KeybState.B = 0
              TurboTimerB = TotalCycles + 800000  ' keep button released for ~1/5s
          END IF
        END IF
      ELSE  ' If turbo B not pressed, check out the regular B button...
        IF (CheckKeyState(KeyDef.B) = 1) THEN ' Button B
            ButtonBisPressed()
          ELSE
            KeybState.B = 0
        END IF
    END IF
  END IF

  JoyNewReg = &b00001111
  SELECT CASE (IoRegisters(&hFF00) AND &b00110000)   ' Get P14 and P15 state
    CASE &b100000  ' P14 selected (bit 4 is low) -> down/up/left/right
      IF KeybState.Down = 1 THEN JoyNewReg = (JoyNewReg AND &b11110111) ' if DOWN -> Set P13
      IF KeybState.Up = 1 THEN JoyNewReg = (JoyNewReg AND &b11111011) ' if UP -> Set P12
      IF KeybState.Left = 1 THEN JoyNewReg = (JoyNewReg AND &b11111101) ' if LEFT -> Set P11
      IF KeybState.Right = 1 THEN JoyNewReg = (JoyNewReg AND &b11111110) ' if RIGHT -> Set P10
    CASE &b010000  ' P15 selected (bit 5 is low) -> Start/Select/B/A
      IF KeybState.Start = 1 THEN JoyNewReg = (JoyNewReg AND &b11110111) ' if ENTER -> Set P13
      IF KeybState.Select = 1 THEN JoyNewReg = (JoyNewReg AND &b11111011) ' if TAB -> Set P12
      IF KeybState.A = 1 THEN JoyNewReg = (JoyNewReg AND &b11111101) ' if CTRL -> Set P11
      IF KeybState.B = 1 THEN JoyNewReg = (JoyNewReg AND &b11111110) ' if ALT -> Set P10
  END SELECT

  IoRegisters(&hFF00) = ((IoRegisters(&hFF00) AND &b11110000) OR JoyNewReg)   ' update the joypad register [FF00h]

  IF (JoyOldReg <> JoyNewReg) THEN                            ' If a key has been pressed/released
    JoyOldReg = JoyNewReg                                     ' update the JoyOldReg variable
  END IF

#ENDMACRO
