REM
REM --------------------------------------
REM This file is part of the zBoy project.
REM Copyright (C) Mateusz Viste 2010
REM --------------------------------------
REM
REM * MBC2 short documentation *
REM
REM This memory controller works much like the MBC1 controller with the following exceptions:
REM MBC2 will work with ROM sizes up to 2Mbit. Writing a value (XXXXBBBB - X = Don't cares, B = bank select
REM bits) into 2000-3FFF area will select an appropriate ROM bank at 4000-7FFF.
REM RAM switching is not provided. Unlike the MBC1 which uses external RAM, MBC2 has 512 x 4 bits of RAM
REM which is in the controller itself. It still requires an external battery to save data during power-off
REM though.
REM The least significant bit of the upper address byte must be zero to enable/disable cart RAM. For example
REM the following addresses can be used to enable/disable cart RAM:
REM 0000-00FF, 0200-02FF, 0400-04FF, ..., 1E00-1EFF.
REM enable/disable is 0000-00FF. The suggested address range to use for MBC2 ram The least significant bit of
REM the upper address byte must be one to select a ROM bank. For example the following addresses can be used
REM to select a ROM bank: 2100-21FF, 2300-23FF, 2500-25FF, ..., 3F00-3FFF. The suggested address range to use
REM for MBC2 rom bank selection is 2100-21FF.
REM
REM Note: The content of the MBC2 RAM can be saved during poweroff if the cartridge comes with a battery.
REM


SUB MBC2_MemoryRead(BYREF memaddr AS UINTEGER, BYREF Result AS UBYTE)
  IF (memaddr <= &h00FF) AND (BootRomEnabledFlag = 1) THEN
    Result = BootROM(memaddr)
  ELSEIF (memaddr >= &hA000) AND (memaddr < &hA200) THEN        ' RAM on cartridge (512x4 bits)
    Result = MemoryBankedRAM(memaddr)
  ELSEIF (memaddr >= &hC000) AND (memaddr < &hE000) THEN    ' Internal 8KiB RAM
    Result = MemoryInternalRAM(memaddr)
  ELSEIF (memaddr >= &hE000) AND (memaddr < &hFE00) THEN    ' RAM mirror
    Result = MemoryInternalRAM(memaddr - 8192)
  ELSEIF (memaddr >= &hFE00) AND (memaddr < &hFEA0) THEN    ' Sprite OAM memory
    Result = SpriteOAM(memaddr)
  ELSEIF (memaddr >= &hFF80) AND (memaddr <= &hFFFF) THEN   ' Hi RAM area
    Result = MemoryInternalHiRAM(memaddr)
  ELSEIF (memaddr >= &h0000) AND (memaddr < &h4000) THEN    ' ROM bank #0
    Result = MemoryROM(memaddr)
  ELSEIF (memaddr >= &hFF00) AND (memaddr <= &hFF4B) THEN   ' I/O registers
    Result = IoRegisters(memaddr)
  ELSEIF (memaddr >= &h4000) AND (memaddr < &h8000) THEN    ' ROM bank #n
    'IF RomInfos.RomSize = 32768 THEN
    '    RETURN MemoryROM(memaddr)
    '  ELSE
        IF (CurRomBank = 0) OR (RomInfos.RomSize = 32768) THEN
            'RETURN MemoryROM(16384 + (memaddr - &h4000))
            Result = MemoryROM(memaddr)
          ELSE
            IF ((CurRomBank SHL 14) + (memaddr - &h4000)) > RomInfos.RomSize THEN                  ' SHL 14 is the same than *16384 (but faster)
              PrintMsg("INVALID ROM ADDRESS!" & CHR(10) & "CURBANK=" & CurRomBank & ")")
              PressAnyKey()
              END
            END IF
            Result = MemoryROM((CurRomBank SHL 14) + (memaddr - &h4000))    ' SHL 14 is the same than *16384 (but faster)
        END IF
    'END IF
  ELSEIF (memaddr >= &h8000) AND (memaddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    Result = VideoRAM(memaddr)
  ELSE                                                  ' Else it's something I don't know about
    'IF (memaddr > &hFFFF) THEN
    '  PrintMsg("MMU FATAL ERROR!" & CHR(10) & "MEMORYREAD() -> INVALID MEMADDR! [" & HEX(memaddr) & "]") : PressAnyKey() : END
    'END IF
    'PRINT "MMU fatal error: Tried to read from an unknown memory address: " & HEX(memaddr) & "h"
    'SLEEP
    'QuitEmulator = 1
    Result = MemoryMAP(memaddr)
  END IF
END SUB


SUB MBC2_MemoryWrite(BYREF memaddr AS UINTEGER, BYREF DataByte AS UBYTE)
  IF (memaddr < &h2000) THEN  ' Enables / disables current RAM bank (not supported yet)
    REM Before using any RAM bank, you need to enable it by writing xxxx1010 into 0000-1FFF (writing there
    REM any non-xxxx1010 value disables the RAM bank).
    IF (DataByte AND &b00001111) = &b00001010 THEN ' Enables currently selected RAM bank
      ELSE  ' Disables currently selected RAM bank
        'UserMessage = "RAM BANK DISABLED"
    END IF
  ELSEIF (memaddr >= &h2000) AND (memaddr < &h4000) THEN       ' Select current ROM bank on MBC1
    CurRomBank = (DataByte AND &b00001111)
  'ELSEIF (memaddr >= &h4000) AND (memaddr < &h8000) THEN
    'SELECT CASE MbcModel ' For mode 16/8, 5 bits is not enough to address all 128 banks (each bank is 16KiB big). Write a
    '  CASE 1 ' byte into 4000-5FFF. The 2 least significant bits of this byte (xxxxxxBB) will be used as the two most
    '    CurRomBank = (CurRomBank AND &b00111111) OR ((DataByte SHL 6) AND &b11000000)  '  significant bits of the ROM address.
    '  CASE 2 ' For mode 4/32, you need to select the RAM bank to be used (at A000-C000). to do that, write a
    '    CurRamBank = (DataByte AND &b00000011)    ' byte into 4000-5FFF. The 2 least significant bits of this
    'END SELECT                                     ' byte (xxxxxxBB) are telling what RAM bank to use.
  'ELSEIF (memaddr >= &h6000) AND (memaddr < &h8000) THEN   ' Configuration for the MBC1 chip (selection of mem model)
  '  IF (DataByte AND &b00000001) = 0 THEN
  '      MbcModel = 1   ' 16/8 model
  '    ELSE
  '      MbcModel = 2   ' 4/32 model
  '  END IF
  ELSEIF (memaddr >= &hA000) AND (memaddr < &hA200) THEN    ' RAM on cartridge (512x4 bits)
    MemoryBankedRAM(memaddr) = DataByte AND &b00001111      ' write only lower 4 bits (this is a 4bit memory)
  ELSEIF (memaddr >= &hC000) AND (memaddr < &hE000) THEN    ' Internal 8KiB RAM
    MemoryInternalRAM(memaddr) = DataByte
  ELSEIF (memaddr >= &hE000) AND (memaddr < &hFE00) THEN    ' RAM mirror
    MemoryInternalRAM(memaddr - 8192) = DataByte
  ELSEIF (memaddr >= &hFE00) AND (memaddr < &hFEA0) THEN    ' Sprite OAM memory
    SpriteOAM(memaddr) = DataByte
  ELSEIF (memaddr >= &hFF80) AND (memaddr <= &hFFFF) THEN   ' Hi RAM area
    MemoryInternalHiRAM(memaddr) = DataByte
  ELSEIF (memaddr = &hFF41) THEN ' STAT register: Do not allow to write into 2 last bits of the STAT register, as these
    IoRegisters(&hFF41) = (IoRegisters(&hFF41) AND &b00000011) OR (DataByte AND &b11111100)    ' bits are the mode flag.
  ELSEIF (memaddr = &hFF44) THEN ' CURLINE [RW] Current Scanline.
    IoRegisters(memaddr) = 0     ' Writing into this register resets it.
    'UserMessage = "LY RESET"
  ELSEIF (memaddr = &hFF46) THEN ' Starts LCD OAM DMA transfer
    STATIC x AS INTEGER          '
    IF DataByte <= &hF1 THEN     ' Normally, should take 160us... but who cares?
        FOR x = 0 TO 159         ' Let's copy XX00-XX9F to FE00-FE9F
          MBC2_MemoryRead((DataByte SHL 8) + x, SpriteOAM(&hFE00 + x))
        NEXT x
      ELSE
        PrintMsg("MMU FATAL ERROR!" & CHR(10) & "TRIED TO RUN LCD OAM DMA TRANSFER USING AN INVALID RANGE!")
        PressAnyKey()
        END
    END IF
  ELSEIF(memaddr = &hFF04) THEN
    IoRegisters(&hFF04) = 0      ' Divide register: Writing any value sets it to 0
  ELSEIF (memaddr >= &hFF00) AND (memaddr <= &hFF4B) THEN   ' I/O registers
    IoRegisters(memaddr) = DataByte
  ELSEIF (memaddr = &hFF50) THEN
    BootRomEnabledFlag = 0
  ELSEIF (memaddr >= &h8000) AND (memaddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    VideoRAM(memaddr) = DataByte
  ELSEIF (memaddr >= &hFEA0) AND (memaddr <= &hFEFF) THEN   ' This area is not usable, but some games writes to it...
    REM                                                     ' For example Dr. Mario tries to write to FEFF... who knows why?
  ELSEIF (memaddr >= &hFF4C) AND (memaddr <= &hFF7F) THEN   ' This area is not usable, but some games writes to it...
    REM                                                     ' For example Dr. Mario tries to write to FF7F... who knows why?
  ELSE                                                      ' Else it's something I don't know about
    'PRINT "MMU fatal error: Tried to write to an unknown memory address: "; HEX(memaddr)
    'QuitEmulator = 1
    MemoryMAP(memaddr) = DataByte
  END IF
END SUB
