REM
REM ----------------------------------------
REM  MMU emulation (Memory Management Unit)
REM  This file is part of the zBoy project.
REM  Copyright (C) Mateusz Viste 2010
REM ----------------------------------------
REM

DIM SHARED MemoryInternalRAM(&hC000 TO &hDFFF) AS UBYTE   ' Internal RAM Area  [8KiB]
DIM SHARED MemoryInternalHiRAM(&hFF80 TO &hFFFF) AS UBYTE ' Internal RAM (high area + IE register)
DIM SHARED MemoryBankedRAM(&hA000 TO &h20A000) AS UBYTE   ' Banked RAM [2MiB]
#IFDEF EMBEDROM
    #INCLUDE EMBEDROM
  #ELSE
    REDIM SHARED MemoryROM(0) AS UBYTE                  ' Declare an empty table for ROM Area (redim later at loadrom time)
#ENDIF
DIM SHARED VideoRAM(&h8000 TO &h9FFF) AS UBYTE      ' Video RAM [8KiB]
DIM SHARED SpriteOAM(&hFE00 TO &hFE9F) AS UBYTE     ' Sprite OAM memory
DIM SHARED IoRegisters(&hFF00 TO &hFF4B) AS UBYTE   ' All I/O memory-mapped registers
DIM SHARED MemoryMAP(&h0000 TO &hFFFF) AS UBYTE     ' Regular memory (fallback for unmapped regions)
DIM SHARED AS UINTEGER CurRamBank = 0    ' Current RAM bank selection
DIM SHARED AS UBYTE Mbc1Model = 1      ' MBC1 memory model (MbcModel can be 1 or 2)  1=16/8 ; 2=4/32
DIM SHARED AS UINTEGER Mbc1RamRomSelect = 0      ' MBC1 RAM or (upper bits) ROM selection register
DIM SHARED AS UINTEGER CurRomBank = 0            ' Used for ROM bank switching (must be at least 9 bits long for MBC5 support!)
DIM SHARED BootRomEnabledFlag AS UBYTE = 0       ' this flag is reseted the first time a write occurs at $FF50
#INCLUDE ONCE "bootrom.bi"                       ' Provides the BootROM(0 TO 255) array

#INCLUDE ONCE "mbc0.bi"  ' Support for ROM-ONLY ROMs
#INCLUDE ONCE "mbc1.bi"  ' Support for MBC1 memory controllers
#INCLUDE ONCE "mbc2.bi"  ' Support for MBC2 memory controllers
#INCLUDE ONCE "mbc3.bi"  ' Support for MBC3 memory controllers (without TIMER support so far)
#INCLUDE ONCE "mbc5.bi"  ' Support for MBC5 memory controllers (usually found in GBC games)


SUB InitRAM() ' Init the RAM areas to random values, as in a real GameBoy
  DIM AS INTEGER x
  FOR x = LBOUND(MemoryInternalRAM) TO UBOUND(MemoryInternalRAM)
    MemoryInternalRAM(x) = INT(RND * 256)
  NEXT x
  FOR x = LBOUND(MemoryInternalHiRAM) TO UBOUND(MemoryInternalHiRAM)
    MemoryInternalHiRAM(x) = INT(RND * 256)
  NEXT x
  FOR x = LBOUND(MemoryBankedRAM) TO UBOUND(MemoryBankedRAM)
    MemoryBankedRAM(x) = INT(RND * 256)
  NEXT x
  FOR x = LBOUND(MemoryMAP) TO UBOUND(MemoryMAP)
    MemoryMAP(x) = INT(RND * 256)
  NEXT x
END SUB


' Below are generic MemoryRead and MemoryWrite routines. These routines
' check if the called address is a well-known address. If this address
' is behaving the same on all known MBC controllers, then it is answered
' here (and it is FAST). Otherwise, a Memory routine specialized for the
' given MBC is called.


#MACRO MemoryRead(ReadAddr, StorageVar)
  IF (ReadAddr <= &h00FF) AND (BootRomEnabledFlag = 1) THEN
    StorageVar = BootROM(ReadAddr)
  ELSEIF (ReadAddr < &h4000) THEN                             ' ROM bank #0
    StorageVar = MemoryROM(ReadAddr)
  ELSEIF (ReadAddr >= &hC000) AND (ReadAddr < &hE000) THEN    ' Internal 8KiB RAM
    StorageVar = MemoryInternalRAM(ReadAddr)
  ELSEIF (ReadAddr >= &hE000) AND (ReadAddr < &hFE00) THEN    ' RAM mirror
    StorageVar = MemoryInternalRAM(ReadAddr - 8192)
  ELSEIF (ReadAddr >= &hFE00) AND (ReadAddr < &hFEA0) THEN    ' Sprite OAM memory
    StorageVar = SpriteOAM(ReadAddr)
  ELSEIF (ReadAddr >= &hFF80) AND (ReadAddr <= &hFFFF) THEN   ' Hi RAM area
    StorageVar = MemoryInternalHiRAM(ReadAddr)
  ELSEIF (ReadAddr >= &hFF00) AND (ReadAddr <= &hFF4B) THEN   ' I/O registers
    StorageVar = IoRegisters(ReadAddr)
  ELSEIF (ReadAddr >= &h8000) AND (ReadAddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    'IF GetLcdMode() = 3 THEN  ' If LCD is in mode 3, return an undefinied value (typically FF)
    '    StorageVar = &hFF
    '  ELSE
        StorageVar = VideoRAM(ReadAddr)
    'END IF
  ELSE
    MemoryReadSpecial(ReadAddr, StorageVar)
  END IF
#ENDMACRO

#MACRO MemoryWrite(WriteAddr, DataHolder)
'SUB MemoryWrite(WriteAddr AS INTEGER, DataHolder AS UBYTE)
  IF (WriteAddr >= &hC000) AND (WriteAddr < &hE000) THEN    ' Internal 8KiB RAM
    IF (TotalCycles < 4000000) THEN ' If the game has just been started (less than 1s) then check if some areas of RAM are not write protected for high-scores loading...
       DIM AS UINTEGER TempUintCounter = 1
       WHILE (SaveScoresWriteProtection(TempUintCounter) <> WriteAddr) AND (SaveScoresWriteProtection(TempUintCounter) <> 0)
         TempUintCounter += 1
       WEND
       IF SaveScoresWriteProtection(TempUintCounter) = 0 THEN MemoryInternalRAM(WriteAddr) = DataHolder
      ELSE
        MemoryInternalRAM(WriteAddr) = DataHolder
    END IF
  ELSEIF (WriteAddr >= &hE000) AND (WriteAddr < &hFE00) THEN    ' RAM mirror
    MemoryInternalRAM(WriteAddr - 8192) = DataHolder
  ELSEIF (WriteAddr >= &hFE00) AND (WriteAddr < &hFEA0) THEN    ' Sprite OAM memory
    SpriteOAM(WriteAddr) = DataHolder
  ELSEIF (WriteAddr >= &hFF80) AND (WriteAddr <= &hFFFF) THEN   ' Hi RAM area
    IF (TotalCycles < 4000000) THEN ' If the game has just been started (less than 1s) then check if some areas of RAM are not write protected for high-scores loading...
       DIM AS UINTEGER TempUintCounter = 1
       WHILE (SaveScoresWriteProtection(TempUintCounter) <> WriteAddr) AND (SaveScoresWriteProtection(TempUintCounter) <> 0)
         TempUintCounter += 1
       WEND
       IF SaveScoresWriteProtection(TempUintCounter) = 0 THEN MemoryInternalHiRAM(WriteAddr) = DataHolder
      ELSE
        MemoryInternalHiRAM(WriteAddr) = DataHolder
    END IF
  ELSEIF (WriteAddr >= &h8000) AND (WriteAddr <= &h9FFF) THEN   ' Video RAM (8KiB)
    'IF GetLcdMode() <> 3 THEN  ' Check if LCD not in mode 3 (if so, write pulse is ignored)
      VideoRAM(WriteAddr) = DataHolder
    'END IF
  ELSE
    MemoryWriteSpecial(WriteAddr, DataHolder)
  END IF
#ENDMACRO
'END SUB


REM Here I am using a trick - declaring FUNCTIONs and SUBs as shared variables, so 
REM I will then assign them a pointer address (depending on the type of memory
REM controller used by the ROM.

DIM SHARED MemoryReadSpecial AS SUB (BYREF memaddr AS UINTEGER, BYREF databyte AS UBYTE)
'FUNCTION MemoryRead(BYREF memaddr AS INTEGER) AS UBYTE
'  SELECT CASE AS CONST RomInfos.MbcModel
'    CASE 0, 1  ' ROM_ONLY or MBC1
'      RETURN MBC1_MemoryRead(memaddr)
'    CASE 5  ' MBC5
'      RETURN MBC5_MemoryRead(memaddr)
'  END SELECT
'END FUNCTION

DIM SHARED MemoryWriteSpecial AS SUB (BYREF memaddr AS UINTEGER, BYREF DataByte AS UBYTE)
'SUB MemoryWrite(BYREF memaddr AS INTEGER, BYREF DataByte AS UBYTE)
'  SELECT CASE AS CONST RomInfos.MbcModel
'    CASE 0, 1  ' ROM_ONLY or MBC1
'      MBC1_MemoryWrite(memaddr, DataByte)
'    CASE 5
'      MBC5_MemoryWrite(memaddr, DataByte)
'  END SELECT
'END SUB
