REM
REM  zBoy - A GameBoy emulator for Windows, Linux and DOS
REM  Copyright (C) Mateusz Viste 2010, 2011
REM

#DEFINE pVer "0.13"
#DEFINE pDate "2010, 2011"

REM  The options below make zBoy to be compiled in debug mode, and optionally trace CPU activity to a trace file
'#DEFINE DEBUGMODE
'#DEFINE DEBUGTRACEFILE
'#DEFINE MEMDUMP_AT_EXIT

REM Enable/disable NetPlay support at compile time
#DEFINE NETPLAYSUPPORT

REM Produce a stand-alone exe file with an embedded ROM file
'#DEFINE EMBEDROM "embedrom/dracula.bi"     ' Embed a ROM at compile time to produce a standalone executable
'#DEFINE EMBEDROM "embedrom/pacman.bi"      ' Embed a ROM at compile time to produce a standalone executable
'#DEFINE EMBEDROM "embedrom/drmario.bi"     ' Embed a ROM at compile time to produce a standalone executable

#IFNDEF DEBUGMODE   ' If not in debugmode, then disable PrintDebug
  #DEFINE PrintDebug REM
#ENDIF

TYPE RomInformations
  Title AS STRING
  NiceTitle AS STRING  ' Same as title, but with lower/upper chars, spaces etc.
  FileName AS STRING   ' The filename of the ROM (without path)
  ColorGB AS BYTE      ' 0 = not color, 1 = color (but still GB compatible), 2 = GBC only (not GB compatible)
  SuperGB AS BYTE      ' 0 = GameBoy, 1 = Super GameBoy
  CartTypeString AS STRING
  CrcSum AS UINTEGER
  MbcModel AS BYTE     ' 0 = ROM ONLY  1 = MBC1  2 = MBC2... -1 = unknown
  Rumble AS BYTE       ' 0 = No rumble  1 = Rumble pak (found only on some MBC5 chips)
  TimerRTC AS BYTE     ' 0 = No RTC  1 = RTC  (found only on some MBC3 chips)
  Battery AS BYTE      ' 0 = no battery, 1 = battery included
  Destination AS BYTE  ' 0 = Japan, 1 = Other
  Licensee AS STRING
  RomSize AS INTEGER
  RamSize AS INTEGER
END TYPE

DIM SHARED RomInfos AS RomInformations
DIM SHARED CartTypes(0 TO 255) AS STRING
DIM SHARED AS BYTE QuitEmulator
DIM SHARED AS UBYTE NetPlay = 0
DIM SHARED AS UBYTE FullscreenFlag = 0, BitDepth = 0, FastScale = 1, GraphicScaleFactor = 4  ' Default Scale is 4x
DIM SHARED AS UINTEGER GraphicWidth, GraphicHeight, ScreenBorderRight = 0, ScreenBorderTop = 0
DIM SHARED AS STRING UserMessage
DIM SHARED AS ULONGINT UserMessageExpirationTime

#INCLUDE ONCE "declares.bi"       ' Some SUB/FUNCTION declarations
#IFDEF NETPLAYSUPPORT
  #INCLUDE ONCE "netplay.bi"        ' Includes network support for 2 player netplay
#ENDIF
#INCLUDE ONCE "numconv.bi"        ' UbyteToByte() and DwordVal()
#INCLUDE ONCE "presskey.bi"       ' PressAnyKey()
#INCLUDE ONCE "datetime.bi"       ' NOW(), YEAR(), MONTH(), DAY(), HOUR(), MINUTE(), etc...
#INCLUDE ONCE "file.bi"           ' Provides the FileExists() function
#INCLUDE ONCE "getsavedir.bi"     ' Provides the GetSaveDir() call
#INCLUDE ONCE "mmu.bi"            ' Include Memory Management Unit emulation core
#INCLUDE ONCE "isdirectory.bi"      ' Provides IsDirectory() (required by save.bi and hiscores.bi)
#INCLUDE ONCE "hiscores.bi"       ' Provides SaveHiScoresMem() and LoadHiScoresMem()
#INCLUDE ONCE "video.bi"          ' Video emulation core
#INCLUDE ONCE "hqvideo.bi"        ' Includes HighQuality output video routines
#INCLUDE ONCE "loadpal.bi"        ' LoadPalette()
#INCLUDE ONCE "adjtiming.bi"      ' AdjustTiming()
#INCLUDE ONCE "joypad.bi"           ' Joypad interface emulation
#INCLUDE ONCE "cpu-z80.bi"          ' Z80 CPU emulation core
#IFDEF DEBUGMODE
  #INCLUDE ONCE "debug.bi"          ' Includes some debugging routines
#ENDIF
#INCLUDE ONCE "battery.bi"          ' Provides LoadBattRAM() and SaveBattRAM()
#INCLUDE ONCE "font.bi"             ' Includes font data used by zBoy to display onscreen messages
#INCLUDE ONCE "wordwrap.bi"         ' WordWrap(), required by PrintMsg()
#INCLUDE ONCE "printmsg.bi"         ' PrintMsg(text AS STRING)
#INCLUDE ONCE "loadrom.bi"          ' LoadROM(RomFile AS STRING)
#INCLUDE ONCE "loadrominfos.bi"     ' LoadRomInfos()
#INCLUDE ONCE "serial.bi"           ' Serial interface emulation
#INCLUDE ONCE "savepcxfile.bi"      ' Screenshot routine SavePcxFile()
#INCLUDE ONCE "chkvideo.bi"         ' ChkVideoMode()
#INCLUDE ONCE "save.bi"             ' SaveGame() and LoadGame()
#INCLUDE ONCE "reset.bi"            ' ResetEmulator()
#INCLUDE ONCE "ssaver.bi"           ' InhibitScreenSaver()

DIM NoYes(0 TO 1) AS STRING => {"No", "Yes"}
DIM AS ULONGINT NextKeybPoll
DIM AS STRING CustomPalette
DIM AS UBYTE x, ShowRomInfos = 0, NoScreenSaver = 0, ShowFPS = 0, HiScoresMem = 0
DIM AS DOUBLE LastDebugOnscreen, OneSecondPollingTimer, LastScreenSaverPoke
#IFNDEF __FB_DOS__
  DIM ScreenSaverPokeThread AS ANY PTR
#ENDIF

REM Here I am preloading a default configuration for input keys mapping
RedefineInputKeys("kb.up","kb.down","kb.left","kb.right","kb.ret","kb.tab","kb.ctrl","kb.alt","none","none")
RedefineFunctionKeys("kb.f1","kb.f2","kb.f5","kb.f7","kb.f8","kb.f9","kb.f10","kb.esc")

RefreshScreen = @RefreshScreen_BasicScale() ' by default, use nearest neighbor scaling

x = 1   ' Here I am parsing command-line parameters
DO
  SELECT CASE UCASE(COMMAND(x))
    CASE "--SCALE"
      x += 1  ' go to next param
      GraphicScaleFactor = VAL(COMMAND(x))  ' retrieve provided scale parameter
      IF (GraphicScaleFactor < 1) OR (GraphicScaleFactor > 7) THEN
          GraphicScaleFactor = 4   ' Fallback to x4
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
        ELSE
          'RefreshScreen = @RefreshScreen_BasicScale() ' Nearest Neighbor (already used by default)
      END IF
    CASE "--FASTSCALE"
      x += 1  ' go to next param
      GraphicScaleFactor = VAL(COMMAND(x))  ' retrieve provided scale parameter
      IF (GraphicScaleFactor < 1) OR (GraphicScaleFactor > 7) THEN
          GraphicScaleFactor = 4   ' Fallback to x4
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
        ELSE
          IF (GraphicScaleFactor > 1) THEN FastScale = 2
      END IF
    CASE "--EPXSCALE"
      x += 1  ' go to next param
      GraphicScaleFactor = VAL(COMMAND(x))  ' retrieve provided scale parameter
      IF (GraphicScaleFactor <> 2) AND (GraphicScaleFactor <> 4) AND (GraphicScaleFactor <> 6) THEN
          GraphicScaleFactor = 4   ' Fallback to x4
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
        ELSE
          RefreshScreen = @RefreshScreen_EPX  ' EPX (aka Scale2x)
      END IF
    CASE "--EPXSCALE3"
      x += 1  ' go to next param
      GraphicScaleFactor = VAL(COMMAND(x))  ' retrieve provided scale parameter
      IF (GraphicScaleFactor <> 3) AND (GraphicScaleFactor <> 6) THEN
          GraphicScaleFactor = 4   ' Fallback to x4
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
        ELSE
          RefreshScreen = @RefreshScreen_EPX3  ' EPX3 (aka AdvMAME3x)
      END IF
    CASE "--EAGLESCALE"
      x += 1  ' go to next param
      GraphicScaleFactor = VAL(COMMAND(x))  ' retrieve provided scale parameter
      IF (GraphicScaleFactor <> 2) AND (GraphicScaleFactor <> 4) AND (GraphicScaleFactor <> 6) THEN
          GraphicScaleFactor = 4   ' Fallback to x4
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
        ELSE
          RefreshScreen = @RefreshScreen_Eagle()  ' Eagle
      END IF
    CASE "--SKIPFRAMES"
      x += 1  ' go to next param
      SkipFrames = VAL(COMMAND(x))  ' retrieve provided scale parameter
      IF (SkipFrames < 1) OR (SkipFrames > 10) THEN
        SkipFrames = 1   ' Fallback to 1 (no frameskip)
        RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
        x = 250  ' make sure to quit parsing now
      END IF
    CASE "--FULLSCREEN"
      FullscreenFlag = 1
    CASE "--BPP"
      x += 1
      BitDepth = VAL(COMMAND(x))
      IF (BitDepth <> 4) AND (BitDepth <> 8) AND (BitDepth <> 16) AND (BitDepth <> 32) THEN
        RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
        x = 250  ' make sure to quit parsing now
        BitDepth = 32
      END IF
    CASE "--BOOTLOGO"  ' Enable Gameboy's bootstrap
      BootRomEnabledFlag = 1
    CASE "--PALETTE"   ' load a custom palette
      x += 1     ' go to next param
      CustomPalette = COMMAND(x)
    CASE "--ROMINFOS"   ' shows various informations about the ROM before running it.
      ShowRomInfos = 1
    CASE "--INPUT"
      IF ReDefineInputKeys(COMMAND(x + 1), COMMAND(x + 2), COMMAND(x + 3), COMMAND(x + 4), COMMAND(x + 5), COMMAND(x + 6), COMMAND(x + 7), COMMAND(x + 8), COMMAND(x + 9), COMMAND(x + 10)) = 0 THEN
          x += 10  ' Redefinition of keys successful, let's continue parsing...
        ELSE  ' If redefining keys failed for some reason, aborting...
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
      END IF
    CASE "--FNKEYS"
      IF ReDefineFunctionKeys(COMMAND(x + 1), COMMAND(x + 2), COMMAND(x + 3), COMMAND(x + 4), COMMAND(x + 5), COMMAND(x + 6), COMMAND(x + 7), COMMAND(x + 8)) = 0 THEN
          x += 8  ' Redefinition of keys successful, let's continue parsing...
        ELSE  ' If redefining keys failed for some reason, aborting...
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
      END IF
    CASE "--NOSPEEDLIMIT"
      NoSpeedLimit = 1
    CASE "--NOSCREENSAVER"
      NoScreenSaver = 1
    CASE "--SHOWFPS"
      ShowFPS = 1
    CASE "--HISCORESMEM"
      HiScoresMem = 1
    CASE "--NETPLAY"
      #IFDEF NETPLAYSUPPORT
        NetPLay = 1
        NetPlayerIP(1) = VAL(COMMAND(x + 1))
        NetPlayerIP(2) = VAL(COMMAND(x + 2))
        NetPlayerIP(3) = VAL(COMMAND(x + 3))
        NetPlayerIP(4) = VAL(COMMAND(x + 4))
        x += 4
      #ELSE
        UserMessage = "NETPLAY NOT AVAILABLE"
      #ENDIF
    CASE ELSE ' anything else is assumed to be a filename
      IF (LEN(RomInfos.Filename) = 0) AND (LEN(COMMAND(x)) > 0) THEN
          RomInfos.Filename = COMMAND(x)
        ELSE
          RomInfos.Filename = ""   ' Clear out filename, to make sure to display usage screen later
          x = 250  ' make sure to quit parsing now
      END IF
  END SELECT
  x += 1
LOOP UNTIL (LEN(COMMAND(x)) = 0) OR (x > 64)


#IFDEF DEBUGMODE
  GraphicWidth = 800
  GraphicHeight = 900
  GraphicScaleFactor = 2
#ELSE
  GraphicWidth = 160 * GraphicScaleFactor
  GraphicHeight = 144 * GraphicScaleFactor
#ENDIF


#IFDEF __FB_DOS__  ' On DOS, make sure to use a standard screen resolution (there's no windowed mode here!)
  FullscreenFlag = 1
#ENDIF

IF BitDepth = 0 THEN  ' If BitDepth hasn't been forced, then choose a default value...
  IF (FullscreenFlag = 1) THEN
      BitDepth = 8   ' default value in fullscreen mode is 8bpp
    ELSE
      BitDepth = 32  ' default value in windowed mode is 32bpp
  END IF
END IF

IF (FullscreenFlag = 1) THEN
    SELECT CASE GraphicScaleFactor
      CASE 1
        GraphicWidth = 320
        GraphicHeight = 200
        ScreenBorderRight = 80  ' add 80 pixels of border on the right side of the screen to horizontally center the display area
        ScreenBorderTop = 27    ' add 27 pixels of border on the top of the screen to vertically center the display area
      CASE 2
        GraphicWidth = 512
        GraphicHeight = 384
        ScreenBorderRight = 96  ' add 96 pixels of border on the right side of the screen to horizontally center the display area
        ScreenBorderTop = 47    ' add 47 pixels of border on the top of the screen to vertically center the display area
      CASE 3
        GraphicWidth = 640
        GraphicHeight = 480
        GraphicScaleFactor = 3
        ScreenBorderRight = 80  ' add 80 pixels of border on the right side of the screen to horizontally center the display area
        ScreenBorderTop = 23    ' add 24 pixels of border on the top of the screen to vertically center the display area
      CASE 4
        GraphicWidth = 800
        GraphicHeight = 600
        ScreenBorderRight = 80  ' add 80 pixels of border on the right side of the screen to horizontally center the display area
        ScreenBorderTop = 11    ' add 11 pixels of border on the top of the screen to vertically center the display area
      CASE 5
        GraphicWidth = 1024
        GraphicHeight = 768
        ScreenBorderRight = 112 ' add 112 pixels of border on the right side of the screen to horizontally center the display area
        ScreenBorderTop = 23    ' add 23 pixels of border on the top of the screen to vertically center the display area
      CASE 6
        GraphicWidth = 1280
        GraphicHeight = 1024
        ScreenBorderRight = 160 ' add 160 pixels of border on the right side of the screen to horizontally center the display area
        ScreenBorderTop = 80    ' add 80 pixels of border on the top of the screen to vertically center the display area
      CASE 7
        GraphicWidth = 1280
        GraphicHeight = 1024
        ScreenBorderRight = 80  ' add 80 pixels of border on the right side of the screen to horizontally center the display area
        ScreenBorderTop = 8     ' add 23 pixels of border on the top of the screen to vertically center the display area
    END SELECT
    REM  fbgfx flags: 0 = Windowed ; 1 = Fullscreen ; 4 = NoSwitch
    IF ChkVideoMode(GraphicWidth, GraphicHeight, BitDepth) = 0 THEN
      PRINT "Resolution not supported by your graphic adapter!"
      END
    END IF
    SCREENRES GraphicWidth, GraphicHeight, BitDepth, 1, 1 ' init a graphic window 1 video page (fullscreen mode)
    SETMOUSE ,,0  ' In fullscreen, hide the mouse cursor
  ELSE
    SCREENRES GraphicWidth, GraphicHeight, BitDepth, 1, 0 ' init a graphic window 1 video page (windowed mode)
    WINDOWTITLE("zBoy GameBoy emulator Copyright (C) Mateusz Viste " & pDate)
    SETMOUSE ,,0  ' hide the mouse cursor
END IF


InitScreenOldBuffer()   ' Init the screen buff used for LCD change detection (will force the complete draw of the 1st frame)
IF (LEN(CustomPalette) > 0) THEN LoadPalette(CustomPalette)  ' Load the palette from given file

IF (BitDepth <= 8) THEN  ' On indexed graphic modes, load the palette into video hardware
  PALETTE 0, ((ScreenPalette32(0) SHR 16) AND &b11111111), ((ScreenPalette32(0) SHR 8) AND &b11111111), ((ScreenPalette32(0)) AND &b11111111)
  PALETTE 1, ((ScreenPalette32(1) SHR 16) AND &b11111111), ((ScreenPalette32(1) SHR 8) AND &b11111111), ((ScreenPalette32(1)) AND &b11111111)
  PALETTE 2, ((ScreenPalette32(2) SHR 16) AND &b11111111), ((ScreenPalette32(2) SHR 8) AND &b11111111), ((ScreenPalette32(2)) AND &b11111111)
  PALETTE 3, ((ScreenPalette32(3) SHR 16) AND &b11111111), ((ScreenPalette32(3) SHR 8) AND &b11111111), ((ScreenPalette32(3)) AND &b11111111)
  PALETTE 14, ((ScreenPalette32(14) SHR 16) AND &b11111111), ((ScreenPalette32(14) SHR 8) AND &b11111111), ((ScreenPalette32(14)) AND &b11111111)
  PALETTE 15, ((ScreenPalette32(15) SHR 16) AND &b11111111), ((ScreenPalette32(15) SHR 8) AND &b11111111), ((ScreenPalette32(15)) AND &b11111111)
  IF (FullscreenFlag = 1) THEN  ' if fullscreen, blank out all the screen to color #15 (0Eh), to get the right color for borders
    LINE (0, 0)-(GraphicWidth - 1, GraphicHeight - 1), 14, BF
  END IF
END IF

ClearScreen()   ' Put color #15 all over the virtual screen


#IFDEF NETPLAYSUPPORT
  DIM NetPollingThread AS ANY PTR
  DIM NetSendingThread AS ANY PTR
  IF (NetPlay = 1) THEN  ' Init net stuff if NetPlay is used
    NetPollingState = 0
    NetPollingThread = ThreadCreate(@NetworkCardPoll)
    NetSendingThread = ThreadCreate(@NetworkCardSend)
  END IF
#ENDIF

#IFNDEF EMBEDROM
IF (LEN(RomInfos.Filename) = 0) THEN
  PrintMsg(CHR(10) & "  *** ZBOY V" & pVer & " ***" & CHR(10,10) & "USAGE: ZBOY FILE.GB" & CHR(10,10) & "SEE ZBOY.TXT FOR DETAILED INSTRUCTIONS.")
  PressAnyKey()
  END
END IF
LoadROM(RomInfos.Filename)
#ENDIF

LoadRomInfos()

IF (ShowRomInfos = 1) THEN
  PrintMsg(UCASE("TITLE: " & RomInfos.NiceTitle & CHR(10,10) & "CRC32: " & HEX(RomInfos.CrcSum, 8) & CHR(10,10) & "LICENSEE: " & RomInfos.Licensee & CHR(10,10) & "CART TYPE: " & RomInfos.CartTypeString & CHR(10,10) & "BATTERY: " & NoYes(RomInfos.Battery) & CHR(10,10) & "ROM SIZE: " & (RomInfos.RomSize \ 1024) & " KIB" & CHR(10,10) & "RAM SIZE: " & (RomInfos.RamSize \ 1024) & " KIB" & CHR(10,10) & "PRESS ANY KEY..."))
'PRINT "Title......: " & RomInfos.Title & " (""" & RomInfos.NiceTitle & """)"
'PRINT "Filename...: " & RomInfos.Filename
'PRINT "CRC32 sum..: " & HEX(RomInfos.CrcSum, 8)
'PRINT "Licensee...: " & RomInfos.Licensee
'IF RomInfos.ColorGB = 2 THEN
'    PRINT "ColorGB....: Yes (GBC ONLY)"
'  ELSE
'    PRINT "ColorGB....: " & NoYes(RomInfos.ColorGB)
'END IF
'PRINT "SuperGB....: " & NoYes(RomInfos.SuperGB)
'PRINT "Cart Type..: " & RomInfos.CartTypeString
'PRINT "Battery....: " & NoYes(RomInfos.Battery)
'IF RomInfos.Destination = 0 THEN
'    PRINT "Country....: Japan"
'  ELSE
'    PRINT "Country....: Other (not Japan)"
'END IF
'PRINT "ROM Size...: " & (RomInfos.RomSize \ 1024) & " KiB"
'PRINT "RAM Size...: " & (RomInfos.RamSize \ 1024) & " KiB"
  SLEEP
END IF

IF RomInfos.Rumble = 1 THEN
  PrintMsg("THIS ROM USES A RUMBLE MECHANISM. ZBOY DOESN'T SUPPORT THAT YET! SORRY.")
  PressAnyKey()
  END
END IF
IF RomInfos.TimerRTC = 1 THEN
  PrintMsg("THIS ROM USES AN INTERNAL RTC TIMER CLOCK. ZBOY DOESN'T SUPPORT SUCH ROMS YET! SORRY.")
  PressAnyKey()
  END
END IF
'IF RomInfos.SuperGB = 1 THEN PRINT : PRINT "Error: SuperGB ROMs are not supported (yet). Sorry." : SLEEP : END
'IF RomInfos.ColorGB = 2 THEN PRINT : PRINT "Error: This ROM requires a GameBoy Color emulator. zBoy supports only legacy GameBoy. Sorry." : SLEEP : END

SELECT CASE RomInfos.MbcModel  ' Check the MBC chip type
  CASE 0  ' ROM ONLY
    MemoryWriteSpecial = @MBC0_MemoryWrite
    MemoryReadSpecial = @MBC0_MemoryRead
  CASE 1  ' MBC1
    MemoryWriteSpecial = @MBC1_MemoryWrite
    MemoryReadSpecial = @MBC1_MemoryRead
  CASE 2  ' MBC2
    MemoryWriteSpecial = @MBC2_MemoryWrite
    MemoryReadSpecial = @MBC2_MemoryRead
  CASE 3  ' MBC3
    MemoryWriteSpecial = @MBC3_MemoryWrite
    MemoryReadSpecial = @MBC3_MemoryRead
  CASE 5  ' MBC5
    MemoryWriteSpecial = @MBC5_MemoryWrite
    MemoryReadSpecial = @MBC5_MemoryRead
  CASE ELSE  ' Any other (not supported) memory controller
    PrintMsg("ERROR: THE MEMORY CONTROLLER USED BY THIS ROM IS NOT SUPPORTED (YET) BY ZBOY. SORRY.")
    PressAnyKey() : END
END SELECT

IF (FullScreenFlag = 0) THEN
  WINDOWTITLE(RomInfos.NiceTitle & " (zBoy)")
END IF

#IFDEF DEBUGMODE
  COLOR RGB(200,200,200)
  InitDebug()
#ENDIF

RANDOMIZE TIMER, 3  ' Use Mersenne Twister RND algorithm for random generation

IF (RomInfos.Battery = 1) AND (RomInfos.RamSize > 0) THEN ' check for battery-saved RAM
  LoadBattRAM()
END IF

'InitRAM()    ' Init RAM areas to random values (as in a real GameBoy)
ResetEmulator() ' Fire up the Zilog80 monster, and init all required variables

IF (HiScoresMem = 1) THEN ' Check out if need to look after hiscores
  LoadHiScoresMem()
END IF

DIM xDone AS UBYTE = 0
DO
  CpuExec()
  CheckSerialLink()  ' Check serial link (not supported yet)
  CheckJoypad()      ' Updates the Joypad register
  CheckInterrupts()  ' Check if any interrupt has been requested since last time
  #IFDEF DEBUGMODE
    IF (TIMER - LastDebugOnscreen) > 0.5 THEN
      DebugSpriteDump(1)  ' perform some debug dumps onscreen (OBJ table at &h8000)
      DebugSpriteDump(2)  ' perform some debug dumps onscreen (OBJ table at &h8800)
      DebugPaletteDump()  ' Dump system's palette onscreen
      DebugShowCPUspeed() ' Calc current CPU speed and put it onscreen
      DebugShowCpuRegisters()  '  Display state of CPU registers
      DebugOnscreen()  ' Display debug buffer informations onscreen
      LastDebugOnscreen = TIMER
    END IF
  #ENDIF

  IF ((TIMER - OneSecondPollingTimer) >= 1) THEN
    OneSecondPollingTimer = TIMER
    IF ShowFPS = 1 THEN
      IF FullScreenFlag = 0 THEN ' if windows mode, display FPS counter in titlebar
          'WINDOWTITLE(RomInfos.NiceTitle & " (zBoy)  /  " & CINT(TotalCycles / ((TIMER - EmulationStartTimer) * 41943.04)) & "%  /  " & CountFPS & " FPS")   ' CPU % & FPS counter
          WINDOWTITLE(RomInfos.NiceTitle & " (zBoy)  /  " & CountFPS & " FPS")   ' FPS counter
        ELSE  ' Else we are in fullscreen, display FPS counter in upper left corner of the screen
          LOCATE 1,1 : PRINT "" & CountFPS & " FPS"
      END IF
    END IF
    CountFPS = 0
    #IFNDEF __FB_DOS__
      IF (NoScreenSaver = 1) THEN
        IF (TIMER - LastScreenSaverPoke) > 55 THEN
          LastScreenSaverPoke = TIMER
          ScreenSaverPokeThread = ThreadCreate(@InhibitScreenSaver, 0) ' run screensaver prevention in separate thread
        END IF                                                         ' to avoid freezes if it takes time to perform.
      END IF
    #ENDIF
  END IF
  IF (LEN(UserMessage) > 0) THEN
    IF (UserMessageExpirationTime = 0) THEN
        UserMessageExpirationTime = TotalCycles + 8000000   ' add about 2sec
      ELSE
        IF (TotalCycles >= UserMessageExpirationTime) THEN
          UserMessageExpirationTime = 0
          UserMessage = ""
        END IF
    END IF
  END IF
  IF (TotalCycles > NextKeybPoll) THEN
    NextKeybPoll += 2000000  ' Poll the keyboard buffer every ~500 ms.
    IF INKEY() = CHR(255) + "k" THEN QuitEmulator = 2 ' Close window signal
  END IF
LOOP UNTIL (QuitEmulator > 0)

IF (RomInfos.Battery = 1) AND (RomInfos.RamSize > 0) THEN ' check for battery-saved RAM
  SaveBattRAM()
END IF

IF (HiScoresMem = 1) THEN
  SaveHiScoresMem()
END IF


#IFDEF MEMDUMP_AT_EXIT
  DIM AS UINTEGER TempUINT             ' -----------------
  OPEN "memdump.bin" FOR OUTPUT AS #1  ' This is a memory
  FOR TempUINT = &h0000 TO &hFFFF      ' dump routine.
    MemoryRead(TempUINT, x)            '     --  --
    PUT #1,, x                         ' Don't include in
  NEXT TempUINT                        ' official release!
  CLOSE #1                             ' -----------------
#ENDIF

#IFDEF DEBUGMODE
  IF (QuitEmulator = 1) THEN
    DebugOnscreen()
    SLEEP
  END IF
#ENDIF

#IFDEF NETPLAYSUPPORT
  IF (NetPlay = 1) THEN  ' Close net stuff if NetPlay is used
    NeedToCloseNetworkThreads = 1
    ThreadWait(NetPollingThread)
    ThreadWait(NetSendingThread)
  END IF
#ENDIF

WHILE LEN(INKEY) > 0 : SLEEP 1,1 : WEND  ' Flush keyb buffer
END
